from ..registers import REG_CTRL1, REG_CTRL2
from ..protocol import BusProtocol


def _set_pga(bus: BusProtocol, addr: int, pga: int) -> None:
    ctrl1 = REG_CTRL1.read(bus, addr)
    if ctrl1.pga != pga:
        ctrl1.pga = pga
        ctrl1.write(bus, addr)


def _set_vldo(bus: BusProtocol, addr: int, vldo: int) -> None:
    ctrl1 = REG_CTRL1.read(bus, addr)
    if ctrl1.vldo != vldo:
        ctrl1.vldo = vldo
        ctrl1.write(bus, addr)


def _set_crs(bus: BusProtocol, addr: int, crs: int) -> None:
    ctrl2 = REG_CTRL2.read(bus, addr)
    if ctrl2.crs != crs:
        ctrl2.crs = crs
        ctrl2.write(bus, addr)


def _set_chs(bus: BusProtocol, addr: int, channel: bool) -> None:
    ctrl2 = REG_CTRL2.read(bus, addr)
    if ctrl2.chs != channel:
        ctrl2.chs = channel
        ctrl2.write(bus, addr)


def _set_cals(bus: BusProtocol, addr: int, cals: bool) -> None:
    ctrl2 = REG_CTRL2.read(bus, addr)
    if ctrl2.cals != cals:
        ctrl2.cals = cals
        ctrl2.write(bus, addr)


def _set_calmod(bus: BusProtocol, addr: int, calmod: int) -> None:
    ctrl2 = REG_CTRL2.read(bus, addr)
    if ctrl2.calmod != calmod:
        ctrl2.calmod = calmod
        ctrl2.write(bus, addr)



from typing import Self
from dataclasses import dataclass

from ..register._byte import ByteRegister


@dataclass
class REG_PU_CTRL(ByteRegister):
    ADDRESS = 0x00

    avdds: bool = False
    oscs: bool = False
    cr: bool = False
    cs: bool = False
    pur: bool = False
    pua: bool = False
    pud: bool = False
    rr: bool = False

    def to_byte(self) -> int:
        value = 0
        value |= int(self.avdds) << 7
        value |= int(self.oscs) << 6
        value |= int(self.cr) << 5
        value |= int(self.cs) << 4
        value |= int(self.pur) << 3
        value |= int(self.pua) << 2
        value |= int(self.pud) << 1
        value |= int(self.rr) << 0
        return value

    @classmethod
    def from_byte(cls, byte: int) -> Self:
        return cls(
            avdds=bool(byte & (1 << 7)),
            oscs=bool(byte & (1 << 6)),
            cr=bool(byte & (1 << 5)),
            cs=bool(byte & (1 << 4)),
            pur=bool(byte & (1 << 3)),
            pua=bool(byte & (1 << 2)),
            pud=bool(byte & (1 << 1)),
            rr=bool(byte & (1 << 0)),
        )



from typing import Literal

from ..protocol import BusProtocol
from .. import registers
from . import power, control, adc


class NAU7802:
    def __init__(self, bus: BusProtocol, addr: int = 0x2A) -> None:
        self.bus = bus
        self.addr: int = addr

    def initialize(self) -> None:
        power._reset(self.bus, self.addr)
        power._power_on(self.bus, self.addr)
        power._set_defaults(self.bus, self.addr)
        adc._set_defaults(self.bus, self.addr)

    def standby(self) -> None:
        power._standby(self.bus, self.addr)

    def resume(self) -> None:
        power._resume(self.bus, self.addr)

    def set_vldo(self, vldo: int) -> None:
        control._set_vldo(self.bus, self.addr, vldo)

    def set_gain(self, gain: int) -> None:
        control._set_pga(self.bus, self.addr, gain)

    def set_crs(self, crs: int) -> None:
        control._set_crs(self.bus, self.addr, crs)

    def set_channel(self, channel: Literal[1, 2]) -> None:
        if channel == 1:
            chs = False
        elif channel == 2:
            chs = True
        else:
            raise RuntimeError(f"Invalid channel {channel}")

        control._set_chs(self.bus, self.addr, chs)

    def set_cals(self, cals: bool) -> None:
        control._set_cals(self.bus, self.addr, cals)

    def set_calmod(self, calmod: int) -> None:
        control._set_calmod(self.bus, self.addr, calmod)

    @property
    def channel(self) -> int:
        return 2 if self.ctrl2.chs else 1

    @property
    def cycle_ready(self) -> bool:
        return self.pu_ctrl.cr

    @property
    def cals(self) -> bool:
        return self.ctrl2.cals

    @property
    def cal_err(self) -> bool:
        return self.ctrl2.cal_err

    @property
    def adc_ctrl1(self) -> registers.REG_ADC_CTRL1:
        return registers.REG_ADC_CTRL1.read(self.bus, self.addr)

    @property
    def adc_ctrl3(self) -> registers.REG_ADC_CTRL3:
        return registers.REG_ADC_CTRL3.read(self.bus, self.addr)

    @property
    def adco(self) -> registers.REG_ADCO:
        return registers.REG_ADCO.read(self.bus, self.addr)

    @property
    def ctrl1(self) -> registers.REG_CTRL1:
        return registers.REG_CTRL1.read(self.bus, self.addr)

    @property
    def ctrl2(self) -> registers.REG_CTRL2:
        return registers.REG_CTRL2.read(self.bus, self.addr)

    @property
    def gcal1(self) -> registers.REG_GCAL1:
        return registers.REG_GCAL1.read(self.bus, self.addr)

    @gcal1.setter
    def gcal1(self, value: int) -> None:
        registers.REG_GCAL1(value=value).write(self.bus, self.addr)

    @property
    def gcal2(self) -> registers.REG_GCAL2:
        return registers.REG_GCAL2.read(self.bus, self.addr)

    @gcal2.setter
    def gcal2(self, value: int) -> None:
        registers.REG_GCAL2(value=value).write(self.bus, self.addr)

    @property
    def i2c_control(self) -> registers.REG_I2C_CONTROL:
        return registers.REG_I2C_CONTROL.read(self.bus, self.addr)

    @property
    def ocal1(self) -> registers.REG_OCAL1:
        return registers.REG_OCAL1.read(self.bus, self.addr)

    @ocal1.setter
    def ocal1(self, value: int) -> None:
        registers.REG_OCAL1(value=value).write(self.bus, self.addr)

    @property
    def ocal2(self) -> registers.REG_OCAL2:
        return registers.REG_OCAL2.read(self.bus, self.addr)

    @ocal2.setter
    def ocal2(self, value: int) -> None:
        registers.REG_OCAL2(value=value).write(self.bus, self.addr)

    @property
    def pu_ctrl(self) -> registers.REG_PU_CTRL:
        return registers.REG_PU_CTRL.read(self.bus, self.addr)

    @property
    def pwr_ctrl(self) -> registers.REG_PWR_CTRL:
        return registers.REG_PWR_CTRL.read(self.bus, self.addr)

    @property
    def rev_id(self) -> registers.REG_REV_ID:
        return registers.REG_REV_ID.read(self.bus, self.addr)



from typing import Self
from dataclasses import dataclass

from ..register._byte import ByteRegister


@dataclass
class REG_PWR_CTRL(ByteRegister):
    ADDRESS = 0x1C

    pga_cap_en: bool = False
    master_bias_curr: int = 0
    adc_curr: int = 0
    pga_curr: int = 0

    def __post_init__(self) -> None:
        if not 0 <= self.master_bias_curr <= 7:
            raise ValueError("master_bias_curr out of range")
        if not 0 <= self.adc_curr <= 3:
            raise ValueError("adc_curr out of range")
        if not 0 <= self.pga_curr <= 3:
            raise ValueError("pga_curr out of range")

    def to_byte(self) -> int:
        value = 0
        value |= int(self.pga_cap_en) << 7
        value |= (self.master_bias_curr & 0b111) << 4
        value |= (self.adc_curr & 0b11) << 2
        value |= self.pga_curr & 0b11
        return value

    @classmethod
    def from_byte(cls, byte: int) -> Self:
        pga_curr = byte & 0b111
        adc_curr = (byte >> 2) & 0b11
        master_bias_curr = (byte >> 4) & 0b111

        return cls(
            pga_cap_en=bool(byte & (1 << 7)),
            master_bias_curr=master_bias_curr,
            adc_curr=adc_curr,
            pga_curr=pga_curr,
        )



from typing import Self
from dataclasses import dataclass

from ..register._byte import ByteRegister


@dataclass
class REG_ADC_CTRL1(ByteRegister):
    ADDRESS = 0x15

    res: int = 0
    reg_chps: int = 0
    adc_vcm: int = 0
    dly_chp: int = 0

    def __post_init__(self) -> None:
        if not 0 <= self.res <= 3:
            raise ValueError("res out of range")
        if not 0 <= self.reg_chps <= 3:
            raise ValueError("reg_chps out of range")
        if not 0 <= self.adc_vcm <= 3:
            raise ValueError("adc_vcm out of range")
        if not 0 <= self.dly_chp <= 3:
            raise ValueError("dly_chp out of range")

    def to_byte(self) -> int:
        value = 0
        value |= (self.res & 0b11) << 6
        value |= (self.reg_chps & 0b11) << 4
        value |= (self.adc_vcm & 0b11) << 2
        value |= self.dly_chp & 0b11
        return value

    @classmethod
    def from_byte(cls, byte: int) -> Self:
        res = (byte >> 6) & 0b11
        reg_chps = (byte >> 4) & 0b11
        adc_vcm = (byte >> 2) & 0b11
        dly_chp = byte & 0b11

        return cls(
            res=res,
            reg_chps=reg_chps,
            adc_vcm=adc_vcm,
            dly_chp=dly_chp,
        )



from typing import Type
from abc import abstractmethod
from dataclasses import dataclass

from ._base import Register, R


@dataclass
class ByteRegister(Register):
    WIDTH = 1

    @abstractmethod
    def to_byte(self) -> int: ...

    @classmethod
    @abstractmethod
    def from_byte(cls: Type[R], data: int) -> R: ...

    def to_bytes(self) -> bytes:
        return bytes([self.to_byte()])

    @classmethod
    def from_bytes(cls: Type[R], data: bytes) -> R:
        cls._assert_data_width(data)
        return cls.from_byte(data[0])



from typing import TypeVar, ClassVar, Type
from abc import ABC, abstractmethod

from ..protocol import BusProtocol

R = TypeVar("R", bound="Register")


class Register(ABC):
    ADDRESS: ClassVar[int]
    WIDTH: int = 1

    @abstractmethod
    def to_bytes(self) -> bytes: ...

    @classmethod
    @abstractmethod
    def from_bytes(cls: Type[R], data: bytes) -> R: ...

    @classmethod
    def from_byte(cls: Type[R], value: int) -> R:
        if cls.WIDTH != 1:
            raise TypeError(f"{cls.__name__} is not 8-bit")
        return cls.from_bytes(bytes([value]))

    @classmethod
    def read(cls: Type[R], bus: BusProtocol, addr: int) -> R:
        if cls.WIDTH == 1:
            raw = bus.read_byte_data(addr, cls.ADDRESS)
            return cls.from_bytes(bytes([raw]))
        else:
            data = bus.read_i2c_block_data(addr, cls.ADDRESS, cls.WIDTH)
            return cls.from_bytes(bytes(data))

    def to_byte(self) -> int:
        if self.WIDTH != 1:
            raise TypeError(f"{type(self).__name__} is not 8-bit")
        return self.to_bytes()[0]

    def write(self, bus: BusProtocol, addr: int) -> None:
        data = self.to_bytes()

        if self.WIDTH == 1:
            bus.write_byte_data(addr, self.ADDRESS, data[0])
        else:
            bus.write_i2c_block_data(addr, self.ADDRESS, list(data))

    @classmethod
    def _assert_data_width(cls: Type[R], data: bytes) -> None:
        if len(data) != cls.WIDTH:
            raise ValueError(
                f"{cls.__name__} expects {cls.WIDTH} bytes, got {len(data)}"
            )



from ..registers import REG_ADC_CTRL1
from ..protocol import BusProtocol


def _set_defaults(bus: BusProtocol, addr: int) -> None:
    adc = REG_ADC_CTRL1.read(bus, addr)
    if adc.reg_chps != 3:
        adc.reg_chps = 3
        adc.write(bus, addr)



from ..registers import REG_PU_CTRL
from ..protocol import BusProtocol


def _reset(bus: BusProtocol, addr: int) -> None:
    pu = REG_PU_CTRL.read(bus, addr)
    pu.rr = True
    pu.write(bus, addr)


def _power_on(bus: BusProtocol, addr: int) -> None:
    pu = REG_PU_CTRL.read(bus, addr)
    pu.rr = False
    pu.pud = True
    pu.write(bus, addr)
    pu = REG_PU_CTRL.read(bus, addr)

    for _ in range(100):
        if REG_PU_CTRL.read(bus, addr).pur:
            return

    raise RuntimeError("Power-up failed")


def _set_defaults(bus: BusProtocol, addr: int) -> None:
    pu = REG_PU_CTRL.read(bus, addr)
    pu.avdds = True
    pu.cr = True
    pu.pua = True
    pu.write(bus, addr)


def _standby(bus: BusProtocol, addr: int) -> None:
    pu = REG_PU_CTRL.read(bus, addr)
    pu.pud = False
    pu.pua = False
    pu.write(bus, addr)


def _resume(bus: BusProtocol, addr: int) -> None:
    pu = REG_PU_CTRL.read(bus, addr)
    pu.pud = True
    pu.pua = True
    pu.write(bus, addr)



from typing import Protocol, List, runtime_checkable


@runtime_checkable
class BusProtocol(Protocol):
    def read_byte_data(self, addr: int, reg: int) -> int: ...
    def write_byte_data(self, addr: int, reg: int, value: int) -> None: ...

    def read_i2c_block_data(self, addr: int, reg: int, size: int) -> list[int]: ...
    def write_i2c_block_data(self, addr: int, reg: int, data: list[int]) -> None: ...
